/*
 * New cryptographic library (ncrypt.dll)
 *
 * Copyright 2016 Alex Henrie
 * Copyright 2020 Ariel Darshan
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include <stdarg.h>

#include "windef.h"
#include "winbase.h"
#include "ncrypt.h"
#include "winreg.h"
#include "wine/debug.h"
#include "wine/heap.h"
#include "ncrypt_internal.h"

WINE_DEFAULT_DEBUG_CHANNEL(ncrypt);

static struct list bufferList = LIST_INIT(bufferList);

SECURITY_STATUS WINAPI NCryptCreatePersistedKey(NCRYPT_PROV_HANDLE provider, NCRYPT_KEY_HANDLE *key,
                                                const WCHAR *algid, const WCHAR *name, DWORD keyspec,
                                                DWORD flags)
{
    struct ncrypt_provider_instance *providerInstance;
    struct ncrypt_key_instance *outKeyInstance;

    TRACE("(0x%lx, %p, %s, %s, 0x%08x, 0x%08x)\n", provider, key, wine_dbgstr_w(algid),
                                                         wine_dbgstr_w(name), keyspec, flags);
    if (!provider)
    {
        return NTE_INVALID_HANDLE;
    }
    providerInstance = handle2provider(provider);
    outKeyInstance = create_key(provider, name);
    if (!outKeyInstance)
    {
        return NTE_NO_MEMORY;
    }
    *key = (NCRYPT_KEY_HANDLE)outKeyInstance;
    return providerInstance->functions.CreatePersistedKey(providerInstance->kspHandle, &outKeyInstance->kspHandle, algid, name, keyspec, flags);
}

SECURITY_STATUS WINAPI NCryptDecrypt(NCRYPT_KEY_HANDLE key, BYTE *input, DWORD insize, void *padding,
                                     BYTE *output, DWORD outsize, DWORD *result, DWORD flags)
{
    struct ncrypt_key_instance *keyInstance;
    struct ncrypt_provider_instance *providerInstance;

    TRACE("(0x%lx, %p, %u, %p, %p, %u, %p, 0x%08x)\n", key, input, insize, padding,
                                                             output, outsize, result, flags);
    if (!key)
    {
        return NTE_INVALID_HANDLE;
    }
    keyInstance = handle2key(key);

    if (!keyInstance->provider)
    {
        return NTE_INVALID_HANDLE;
    }
    providerInstance = handle2provider(keyInstance->provider);

    return providerInstance->functions.Decrypt(providerInstance->kspHandle, keyInstance->kspHandle, input, insize, padding, output, outsize, result, flags);
}

SECURITY_STATUS WINAPI NCryptDeleteKey(NCRYPT_KEY_HANDLE key, DWORD flags)
{
    struct ncrypt_key_instance *keyInstance;
    struct ncrypt_provider_instance *providerInstance;
    SECURITY_STATUS ret;

    TRACE("(0x%lx, 0x%08x)", key, flags);

    if (!key)
    {
        return NTE_INVALID_HANDLE;
    }
    keyInstance = handle2key(key);

    if (!keyInstance->provider)
    {
        return NTE_INVALID_HANDLE;
    }
    providerInstance = handle2provider(keyInstance->provider);

    ret = providerInstance->functions.DeleteKey(providerInstance->kspHandle, keyInstance->kspHandle, flags);
    if (ret != ERROR_SUCCESS)
    {
        return ret;
    }
    keyInstance->kspHandle = 0;

    free_key(key);

    return ret;
}

SECURITY_STATUS WINAPI NCryptEncrypt(NCRYPT_KEY_HANDLE key, BYTE *input, DWORD insize, void *padding,
                                     BYTE *output, DWORD outsize, DWORD *result, DWORD flags)
{
    struct ncrypt_key_instance *keyInstance;
    struct ncrypt_provider_instance *providerInstance;

    TRACE("(0x%lx, %p, %u, %p, %p, %u, %p, 0x%08x)\n", key, input, insize, padding,
                                                             output, outsize, result, flags);

    if (!key)
    {
        return NTE_INVALID_HANDLE;
    }
    keyInstance = handle2key(key);

    if (!keyInstance->provider)
    {
        return NTE_INVALID_HANDLE;
    }
    providerInstance = handle2provider(keyInstance->provider);

    return providerInstance->functions.Encrypt(providerInstance->kspHandle, keyInstance->kspHandle, input, insize, padding, output, outsize, result, flags);
}

SECURITY_STATUS WINAPI NCryptEnumAlgorithms(NCRYPT_PROV_HANDLE provider, DWORD alg_ops,
                                            DWORD *alg_count, NCryptAlgorithmName **alg_list,
                                            DWORD flags)
{
    struct ncrypt_provider_instance *providerInstance;
    SECURITY_STATUS ret;

    TRACE("(0x%lx, 0x%08x, %p, %p, 0x%08x)\n", provider, alg_ops, alg_count, alg_list, flags);

    if (!provider)
    {
        return NTE_INVALID_HANDLE;
    }
    providerInstance = handle2provider(provider);
    ret = providerInstance->functions.EnumAlgorithms(providerInstance->kspHandle, alg_ops, alg_count, alg_list, flags);
    if (ret == ERROR_SUCCESS)
    {
        register_buffer(provider, *alg_list);
    }

    return ret;
}

SECURITY_STATUS WINAPI NCryptEnumKeys(NCRYPT_PROV_HANDLE provider, const WCHAR *scope,
                                      NCryptKeyName **key_name, PVOID *enum_state, DWORD flags)
{
    struct ncrypt_provider_instance *providerInstance;
    SECURITY_STATUS ret;
    BOOL firstRound;

    TRACE("(0x%lx, %p, %p, %p, 0x%08x)\n", provider, scope, key_name, enum_state, flags);

    if (!provider)
    {
        return NTE_INVALID_HANDLE;
    }
    providerInstance = handle2provider(provider);

    firstRound = (*enum_state == NULL);

    ret = providerInstance->functions.EnumKeys(providerInstance->kspHandle, scope, key_name, enum_state, flags);

    if (firstRound && *enum_state)
    {
        ret = register_buffer(provider, *enum_state);
        if (ret == NTE_NO_MEMORY)
        {
            ret = providerInstance->functions.FreeBuffer(*enum_state);
            if (ret != ERROR_SUCCESS)
            {
                ERR("Provider failed to free enum_state\n");
            }
            *enum_state = NULL;

            if (*key_name)
            {
                ret = providerInstance->functions.FreeBuffer(*key_name);
                if (ret != ERROR_SUCCESS)
                {
                    ERR("Provider failed to free key_name\n");
                }
            }
            *key_name = NULL;
            return NTE_NO_MEMORY;
        }
        else if (ret != ERROR_SUCCESS)
        {
            FIXME("Unknown error while registering buffer: 0x%08x", ret);
        }
    }

    if (ret == ERROR_SUCCESS && *key_name)
    {
        ret = register_buffer(provider, *key_name);
        if (ret == NTE_NO_MEMORY)
        {
            ret = providerInstance->functions.FreeBuffer(*key_name);
            if (ret != ERROR_SUCCESS)
            {
                ERR("Provider failed to free key_name\n");
            }
            *key_name = NULL;

            if (*enum_state)
            {
                ret = NCryptFreeBuffer(*enum_state);
                if (ret != ERROR_SUCCESS)
                {
                    ERR("Provider failed to unregister enum_state\n");
                }
            }
            *enum_state = NULL;
            return NTE_NO_MEMORY;
        }
        else if (ret != ERROR_SUCCESS)
        {
            FIXME("Unknown error while registering buffer: 0x%08x", ret);
        }
    }
    return ret;
}

SECURITY_STATUS WINAPI NCryptEnumStorageProviders(DWORD *providerCount, NCryptProviderName **providerList, DWORD flags)
{
    HKEY hKey;
    HKEY hSubKey;
    SECURITY_STATUS ret;
    LSTATUS status;
    DWORD maxKeyLen;
    DWORD i;
    WCHAR providerName[MAX_PATH];
    WCHAR providerKSReg[MAX_PATH];
    DWORD providerNameLength;
    NCryptProviderName *outList;
    DWORD outCount;
    DWORD numKeys;
    void *allocBuffer;
    void *reAllocBuffer;
    DWORD allocSize;
    DWORD totalNamesSize;

    TRACE("(%p, %p, 0x%08x)\n", providerCount, providerList, flags);

    if (flags)
    {
        FIXME("Ignoring flags: 0x%08x - Not implemented\n", flags);
    }

    ret = ERROR_SUCCESS;
    outList = NULL;
    outCount = 0;
    allocSize = 0;

    status = RegOpenKeyExW(HKEY_LOCAL_MACHINE, L"SYSTEM\\CurrentControlSet\\Control\\Cryptography\\Providers", 0, KEY_READ, &hKey);
    if (status != ERROR_SUCCESS)
    {
        ERR("Failed to open 'HKLM\\SYSTEM\\CurrentControlSet\\Control\\Cryptography\\Providers' error code: 0x%08x\n", status);
        return NTE_NOT_SUPPORTED; /* FIXME: what should be returned here? */
    }

    status = RegQueryInfoKeyW(hKey, NULL, NULL, NULL, &numKeys, &maxKeyLen, NULL, NULL, NULL, NULL, NULL, NULL);
    if (status != ERROR_SUCCESS)
    {
        ERR("Failed to query info on 'HKLM\\SYSTEM\\CurrentControlSet\\Control\\Cryptography\\Providers' error code: 0x%08x\n", status);
        ret = NTE_NOT_SUPPORTED; /* FIXME: what should be returned here? */
        goto cleanup;
    }


    for (i = 0; i < numKeys; ++i)
    {
        providerNameLength = MAX_PATH;
        status = RegEnumKeyExW(hKey, i, providerName, &providerNameLength, NULL, NULL, NULL, NULL);
        if (status != ERROR_SUCCESS)
        {
            ERR("Failed to enumerate sub keys of 'HKLM\\SYSTEM\\CurrentControlSet\\Control\\Cryptography\\Providers' error code: 0x%08x\n", status);
            ret = NTE_NOT_SUPPORTED; /* FIXME: what should be returned here? */
            goto cleanup;
        }


        /* This registry key should exist if the provider supports key storage */
        swprintf(providerKSReg, MAX_PATH - 1, L"%s\\UM\\00010001", providerName);
        status = RegOpenKeyExW(hKey, providerKSReg, 0, KEY_READ, &hSubKey);
        if (status != ERROR_SUCCESS)
        {
            /* Provider doesn't support key storage, skip it */
            continue;
        }
        RegCloseKey(hSubKey);

        allocSize += sizeof(NCryptProviderName) + (providerNameLength + 1) * sizeof(WCHAR);
        if (!outCount++)
        {
            allocBuffer = heap_alloc(allocSize);
            if (!allocBuffer)
            {
                ERR("Out of memory!\n");
                ret = NTE_NO_MEMORY;
                goto cleanup;
            }
            totalNamesSize = 0;
        }
        else
        {
            reAllocBuffer = heap_realloc(allocBuffer, allocSize);
            if (!reAllocBuffer)
            {
                ERR("Out of memory during realloc!\n");
                heap_free(allocBuffer);
                ret = NTE_NO_MEMORY;
                goto cleanup;
            }
            /* Shift the wstrings that contain the names over by the size of one struct */
            memmove((BYTE*)allocBuffer + outCount * sizeof(NCryptProviderName), (BYTE*)allocBuffer + (outCount - 1) * sizeof(NCryptProviderName), totalNamesSize);
        }



        outList = allocBuffer;
        outList[outCount - 1].pszName = (WCHAR*)((BYTE*)allocBuffer + outCount * sizeof(NCryptProviderName) + totalNamesSize);
        outList[outCount - 1].pszComment = NULL;

        wcscpy(outList[outCount - 1].pszName, providerName);
        totalNamesSize += (providerNameLength + 1) * sizeof(WCHAR);
    }


cleanup:
    if (ret == ERROR_SUCCESS)
    {
        *providerCount = outCount;
        *providerList = outList;
        register_buffer(0, outList);
    }

    RegCloseKey(hKey);
    return ret;
}

SECURITY_STATUS WINAPI NCryptExportKey(NCRYPT_PROV_HANDLE provider, NCRYPT_KEY_HANDLE enc_key, const WCHAR *blobType,
                                       NCryptBufferDesc *parameterList, PBYTE output, DWORD outsize, DWORD *result, DWORD flags)
{
    struct ncrypt_provider_instance *providerInstance;
    struct ncrypt_key_instance *encryptKeyInstance;
    NCRYPT_KEY_HANDLE kspEncryptKeyHandle;

    TRACE("(0x%lx, 0x%lx, %s, %p, %p, 0x%08x, %p, 0x%08x)\n", provider, enc_key, wine_dbgstr_w(blobType),
                                                                parameterList, output, outsize, result, flags);

    if (!provider)
    {
        return NTE_INVALID_HANDLE;
    }
    providerInstance = handle2provider(provider);

    kspEncryptKeyHandle = 0;
    if (enc_key)
    {
        encryptKeyInstance = handle2key(enc_key);
        kspEncryptKeyHandle = encryptKeyInstance->kspHandle;
    }

    return providerInstance->functions.ExportKey(providerInstance->kspHandle, kspEncryptKeyHandle, blobType, parameterList, output, outsize, result, flags);
}

SECURITY_STATUS WINAPI NCryptFinalizeKey(NCRYPT_KEY_HANDLE key, DWORD flags)
{
    struct ncrypt_key_instance *keyInstance;
    struct ncrypt_provider_instance *providerInstance;

    TRACE("(0x%lx, 0x%08x)\n", key, flags);


    if (!key)
    {
        return NTE_INVALID_HANDLE;
    }
    keyInstance = handle2key(key);

    if (!keyInstance->provider)
    {
        return NTE_INVALID_HANDLE;
    }
    providerInstance = handle2provider(keyInstance->provider);

    return providerInstance->functions.FinalizeKey(providerInstance->kspHandle, keyInstance->kspHandle, flags);
}

SECURITY_STATUS WINAPI NCryptFreeBuffer(PVOID buf)
{
    struct ncrypt_buffer *cursor, *safetyCursor;
    struct ncrypt_provider_instance *providerInstance;
    SECURITY_STATUS ret;
    BOOL found;

    TRACE("(%p)\n", buf);

    found = FALSE;

    LIST_FOR_EACH_ENTRY_SAFE(cursor, safetyCursor, &bufferList, struct ncrypt_buffer, entry)
    {
        if (cursor->buffer == buf)
        {
            found = TRUE;
            list_remove(&cursor->entry);

            if (!cursor->owner)
            {
                heap_free(buf);
                heap_free(cursor);
                break;
            }

            providerInstance = handle2provider(cursor->owner);
            ret = providerInstance->functions.FreeBuffer(buf);
            if (ret != ERROR_SUCCESS)
            {
                ERR("Provider failed to free buffer: %p\n", buf);
                return ret;
            }
            unref_provider(cursor->owner);
            heap_free(cursor);
            break;
        }
    }

    if (!found)
    {
        WARN("Couldn't find buffer in list of registered buffers\n");
        return NTE_INVALID_PARAMETER;
    }
    return ERROR_SUCCESS;
}

SECURITY_STATUS WINAPI NCryptFreeObject(NCRYPT_HANDLE object)
{
    /* FIXME: implement for secret and hash handles */
    struct ncrypt_base_object *obj;

    TRACE("(0x%lx)\n", object);

    obj = handle2baseObject(object);

    switch(obj->sType)
    {
        case NCRYPT_OBJ_TYPE_PROVIDER:
            return free_provider(object);

        case NCRYPT_OBJ_TYPE_KEY:
            return free_key(object);

        default:
            FIXME("Object type not implemented: 0x%08x\n", obj->sType);
            return NTE_NOT_SUPPORTED;
    }
}

SECURITY_STATUS WINAPI NCryptGetProperty(NCRYPT_HANDLE object, const WCHAR *property, PBYTE output,
                                         DWORD outsize, DWORD *result, DWORD flags)
{
    struct ncrypt_base_object *obj;
    struct ncrypt_key_instance *keyInstance;
    struct ncrypt_provider_instance *providerInstance;

    TRACE("(0x%lx, %s, %p, %u, %p, 0x%08x)\n", object, wine_dbgstr_w(property), output, outsize,
                                                         result, flags);

    if (!object)
    {
        return NTE_INVALID_HANDLE;
    }

    obj = handle2baseObject(object);

    switch(obj->sType)
    {
        case NCRYPT_OBJ_TYPE_PROVIDER:
            providerInstance = handle2provider(object);
            return providerInstance->functions.GetProviderProperty(providerInstance->kspHandle, property, output, outsize, result, flags);

        case NCRYPT_OBJ_TYPE_KEY:
            keyInstance = handle2key(object);
            providerInstance = handle2provider(keyInstance->provider);
            return providerInstance->functions.GetKeyProperty(providerInstance->kspHandle, keyInstance->kspHandle, property, output, outsize, result, flags);

        default:
            FIXME("Object type not implemented: 0x%08x\n", obj->sType);
            return NTE_NOT_SUPPORTED;
    }
}

SECURITY_STATUS WINAPI NCryptImportKey(NCRYPT_PROV_HANDLE provider, NCRYPT_KEY_HANDLE decrypt_key,
                                       const WCHAR *type, NCryptBufferDesc *params, NCRYPT_KEY_HANDLE *key,
                                       PBYTE data, DWORD datasize, DWORD flags)
{
    struct ncrypt_provider_instance *providerInstance;
    struct ncrypt_key_instance *decryptKeyInstance;
    struct ncrypt_key_instance *outKeyInstance;
    NCRYPT_KEY_HANDLE kspDecryptKeyHandle;

    TRACE("(0x%lx, 0x%lx, %s, %p, %p, %p, %u, 0x%08x)\n", provider, decrypt_key,
                                                                wine_dbgstr_w(type), params,
                                                                key, data, datasize, flags);
    if (!provider)
    {
        return NTE_INVALID_HANDLE;
    }
    providerInstance = handle2provider(provider);

    kspDecryptKeyHandle = 0;
    if (decrypt_key)
    {
        decryptKeyInstance = handle2key(decrypt_key);
        kspDecryptKeyHandle = decryptKeyInstance->kspHandle;
    }

    outKeyInstance = create_key(provider, NULL);
    if (!outKeyInstance)
    {
        return NTE_NO_MEMORY;
    }
    *key = (NCRYPT_KEY_HANDLE)outKeyInstance;
    return providerInstance->functions.ImportKey(providerInstance->kspHandle, kspDecryptKeyHandle, type, params, &outKeyInstance->kspHandle, data, datasize, flags);
}

SECURITY_STATUS WINAPI NCryptIsAlgSupported(NCRYPT_PROV_HANDLE provider, const WCHAR *algid,
                                            DWORD flags)
{
    struct ncrypt_provider_instance *providerInstance;

    TRACE("(0x%lx, %s, 0x%08x)\n", provider, wine_dbgstr_w(algid), flags);

    if (!provider)
    {
        return NTE_INVALID_HANDLE;
    }
    providerInstance = handle2provider(provider);
    return providerInstance->functions.IsAlgSupported(providerInstance->kspHandle, algid, flags);
}

BOOL WINAPI NCryptIsKeyHandle(NCRYPT_KEY_HANDLE hKey)
{
    struct ncrypt_base_object *obj;

    TRACE("(0x%lx)\n", hKey);

    obj = handle2baseObject(hKey);
    return obj->sType == NCRYPT_OBJ_TYPE_KEY;
}

SECURITY_STATUS WINAPI NCryptOpenKey(NCRYPT_PROV_HANDLE provider, NCRYPT_KEY_HANDLE *key,
                                     const WCHAR *name, DWORD keyspec, DWORD flags)
{
    struct ncrypt_provider_instance *providerInstance;
    struct ncrypt_key_instance *outKeyInstance;

    TRACE("(0x%lx, %p, %s, 0x%08x, 0x%08x)\n", provider, key, wine_dbgstr_w(name), keyspec, flags);

    if (!provider)
    {
        return NTE_INVALID_HANDLE;
    }
    providerInstance = handle2provider(provider);

    outKeyInstance = create_key(provider, name);
    if (!outKeyInstance)
    {
        return NTE_NO_MEMORY;
    }
    *key = (NCRYPT_KEY_HANDLE)outKeyInstance;

    return providerInstance->functions.OpenKey(providerInstance->kspHandle, &outKeyInstance->kspHandle, name, keyspec, flags);
}

SECURITY_STATUS WINAPI NCryptOpenStorageProvider(NCRYPT_PROV_HANDLE *provider, const WCHAR *name, DWORD flags)
{
    NCryptProviderName *providerNames;
    DWORD providerCount;
    DWORD i;
    BOOL isKSP;
    LSTATUS lstatus;
    SECURITY_STATUS secStatus;
    WCHAR registryPath[MAX_PATH];
    WCHAR dllName[MAX_PATH];
    DWORD dllNameLen;

    TRACE("(%p, %s, %u)\n", provider, wine_dbgstr_w(name), flags);


    secStatus = NCryptEnumStorageProviders(&providerCount, &providerNames, 0);
    if (secStatus != ERROR_SUCCESS)
    {
        return secStatus;
    }

    isKSP = FALSE;
    for (i = 0; i < providerCount; ++i)
    {
        if (!wcscmp(providerNames[i].pszName, name))
        {
            isKSP = TRUE;
            break;
        }
    }

    secStatus = NCryptFreeBuffer(providerNames);
    if (secStatus != ERROR_SUCCESS)
    {
        WARN("Failed to free providers list\n");
    }

    if (!isKSP)
    {
        ERR("provider not found: %s\n", wine_dbgstr_w(name));
        return NTE_INVALID_PARAMETER;
    }



    dllNameLen = MAX_PATH;
    swprintf(registryPath, MAX_PATH, L"SYSTEM\\CurrentControlSet\\Control\\Cryptography\\Providers\\%s\\UM", name);

    lstatus = RegGetValueW(HKEY_LOCAL_MACHINE, registryPath, L"Image", RRF_RT_REG_SZ, NULL, dllName, &dllNameLen);
    if (lstatus != ERROR_SUCCESS)
    {
        ERR("Couldn't find image for provider: %s\n", wine_dbgstr_w(name));
        return NTE_INVALID_PARAMETER;
    }

    return open_provider(provider, name, dllName, flags);
}

SECURITY_STATUS WINAPI NCryptSetProperty(NCRYPT_HANDLE object, const WCHAR *property,
                                         PBYTE input, DWORD insize, DWORD flags)
{
    struct ncrypt_base_object *obj;
    struct ncrypt_key_instance *keyInstance;
    struct ncrypt_provider_instance *providerInstance;

    TRACE("(%lx, %s, %p, %u, 0x%08x)\n", object, wine_dbgstr_w(property), input, insize,
                                               flags);
    if (!object)
    {
        return NTE_INVALID_HANDLE;
    }

    obj = handle2baseObject(object);

    switch(obj->sType)
    {
        case NCRYPT_OBJ_TYPE_PROVIDER:
            providerInstance = handle2provider(object);
            return providerInstance->functions.SetProviderProperty(providerInstance->kspHandle, property, input, insize, flags);

        case NCRYPT_OBJ_TYPE_KEY:
            keyInstance = handle2key(object);
            providerInstance = handle2provider(keyInstance->provider);
            return providerInstance->functions.SetKeyProperty(providerInstance->kspHandle, keyInstance->kspHandle, property, input, insize, flags);

        default:
            FIXME("Object type not implemented: 0x%08x\n", obj->sType);
            return NTE_NOT_SUPPORTED;
    }
}

SECURITY_STATUS WINAPI NCryptSignHash(NCRYPT_KEY_HANDLE key, void *paddingInfo, PBYTE hashValue, DWORD hashSize,
                                      PBYTE signature, DWORD signatureSize, DWORD *result, DWORD flags)
{
    struct ncrypt_key_instance *keyInstance;
    struct ncrypt_provider_instance * providerInstance;

    TRACE("(0x%lx, %p, %p, 0x%08x, %p, 0x%08x, %p, 0x%08x)\n", key, paddingInfo, hashValue, hashSize,
                                                                 signature, signatureSize, result, flags);

    if (!key)
    {
        return NTE_INVALID_HANDLE;
    }
    keyInstance = handle2key(key);

    if (!keyInstance->provider)
    {
        return NTE_INVALID_HANDLE;
    }
    providerInstance = handle2provider(keyInstance->provider);

    return providerInstance->functions.SignHash(keyInstance->kspHandle, paddingInfo, hashValue, hashSize, signature, signatureSize, result, flags);
}

SECURITY_STATUS WINAPI NCryptVerifySignature(NCRYPT_KEY_HANDLE key, void *paddingInfo, PBYTE hashValue,
                                             DWORD hashSize, PBYTE signature, DWORD signatureSize, DWORD flags)
{
    struct ncrypt_key_instance *keyInstance;
    struct ncrypt_provider_instance * providerInstance;

    TRACE("(0x%lx, %p, %p, 0x%08x, %p, 0x%08x, 0x%08x)\n", key, paddingInfo, hashValue, hashSize,
                                                            signature, signatureSize, flags);

    if (!key)
    {
        return NTE_INVALID_HANDLE;
    }
    keyInstance = handle2key(key);

    if (!keyInstance->provider)
    {
        return NTE_INVALID_HANDLE;
    }
    providerInstance = handle2provider(keyInstance->provider);

    return providerInstance->functions.VerifySignature(keyInstance->kspHandle, paddingInfo, hashValue, hashSize, signature, signatureSize, flags);
}

static SECURITY_STATUS open_provider(NCRYPT_PROV_HANDLE *providerInstance, const WCHAR *name, const WCHAR *dllName, DWORD flags)
{
    unsigned int size;
    void *buffer;
    struct ncrypt_provider_instance *ret;
    HMODULE hDLL;
    getInterfaceFn getKeyStorageInterface;
    SECURITY_STATUS status;
    NCRYPT_KEY_STORAGE_FUNCTION_TABLE *functionTable;
    NCRYPT_PROV_HANDLE kspHandle;

    hDLL = LoadLibraryW(dllName);
    if (!hDLL)
    {
        ERR("Failed to load provider dll: %s. error code: 0x%08x\n", wine_dbgstr_w(dllName), GetLastError());
        return NTE_PROV_DLL_NOT_FOUND;
    }

    getKeyStorageInterface = (getInterfaceFn) GetProcAddress(hDLL, "GetKeyStorageInterface");
    if (!getKeyStorageInterface)
    {
        ERR("Failed to locate the 'GetKeyStorageInterface' export in dll: %s. error code: 0x%08x\n", wine_dbgstr_w(dllName), GetLastError());
        FreeLibrary(hDLL);
        return NTE_PROVIDER_DLL_FAIL;
    }

    /* FIXME: What should flags be? */
    status = getKeyStorageInterface(name, &functionTable, 0);
    if (status != ERROR_SUCCESS)
    {
        ERR("Failed to get key storage interface from dll: %s\n", wine_dbgstr_w(dllName));
        FreeLibrary(hDLL);
        return status;
    }

    status = functionTable->OpenProvider(&kspHandle, name, flags);
    if (status != ERROR_SUCCESS)
    {
        ERR("OpenProvider failed in dll: %s\n", wine_dbgstr_w(dllName));
        FreeLibrary(hDLL);
        return status;
    }

    /* Allocate both structure and name in the same call */
    size = sizeof(struct ncrypt_provider_instance) + (wcslen(name) + 1) * sizeof(WCHAR);

    buffer = heap_alloc(size);
    if (!buffer)
    {
        ERR("Out of memory!\n");
        FreeLibrary(hDLL);
        return NTE_NO_MEMORY;
    }

    ret = (struct ncrypt_provider_instance*)buffer;
    ret->sType = NCRYPT_OBJ_TYPE_PROVIDER;
    ret->refCount = 1;
    ret->kspDLL = hDLL;
    memcpy(&ret->functions, functionTable, sizeof(NCRYPT_KEY_STORAGE_FUNCTION_TABLE));
    ret->kspHandle = kspHandle;
    ret->name = (WCHAR*)((char*)buffer + sizeof(struct ncrypt_provider_instance));
    wcscpy(ret->name, name);

    *providerInstance = provider2handle(ret);
    return ERROR_SUCCESS;
}

static struct ncrypt_key_instance* create_key(NCRYPT_PROV_HANDLE provider, const WCHAR *name)
{
    unsigned int size;
    void *buffer;
    struct ncrypt_key_instance *ret;

    size = sizeof(struct ncrypt_key_instance);

    if (name)
    {
        /* Allocate both structure and name in the same call */
        size += (wcslen(name) + 1) * sizeof(WCHAR);
    }

    buffer = heap_alloc(size);
    if (!buffer)
    {
        return NULL;
    }

    ret = (struct ncrypt_key_instance *) buffer;
    ret->sType = NCRYPT_OBJ_TYPE_KEY;
    ret->provider = ref_provider(provider);

    if (name)
    {
        ret->name = (WCHAR*)((char*)buffer + sizeof(struct ncrypt_key_instance));
        wcscpy(ret->name, name);
    }
    else
    {
        ret->name = NULL;
    }

    return ret;
}

static SECURITY_STATUS register_buffer(NCRYPT_PROV_HANDLE provider, const void *buffer)
{
    struct ncrypt_buffer *bufferEntry;

    bufferEntry = heap_alloc(sizeof(struct ncrypt_buffer));
    if (!bufferEntry)
    {
        return NTE_NO_MEMORY;
    }

    bufferEntry->buffer = buffer;
    bufferEntry->owner = ref_provider(provider);

    list_add_head(&bufferList, &bufferEntry->entry);
    return ERROR_SUCCESS;
}

static NCRYPT_PROV_HANDLE ref_provider(NCRYPT_PROV_HANDLE provider)
{
    struct ncrypt_provider_instance *providerInstance;

    if (provider)
    {
        providerInstance = handle2provider(provider);
        providerInstance->refCount++;
    }

    return provider;
}

static void unref_provider(NCRYPT_PROV_HANDLE provider)
{
    struct ncrypt_provider_instance *providerInstance;

    if (provider)
    {
        providerInstance = handle2provider(provider);
        providerInstance->refCount--;
    }
}

static SECURITY_STATUS free_provider(NCRYPT_HANDLE provider)
{
    struct ncrypt_provider_instance *providerInstance;
    SECURITY_STATUS ret;
    BOOL success;

    if (!provider)
    {
        return NTE_INVALID_HANDLE;
    }

    providerInstance = handle2provider(provider);

    if (providerInstance->refCount)
    {
        providerInstance->refCount--;
    }

    if (!providerInstance->refCount)
    {
        if (providerInstance->kspHandle)
        {
            ret = providerInstance->functions.FreeProvider(providerInstance->kspHandle);
            if (ret != ERROR_SUCCESS)
            {
                ERR("FreeProvider failed\n");
                return ret;
            }
            providerInstance->kspHandle = 0;
        }

        if (providerInstance->kspDLL)
        {
            success = FreeLibrary(providerInstance->kspDLL);
            if (!success)
            {
                ERR("Failed to free dll: 0x%08x\n", GetLastError());
                return NTE_INTERNAL_ERROR;
            }
            providerInstance->kspDLL = NULL;
        }
        heap_free(providerInstance);
    }
    return ERROR_SUCCESS;
}

static SECURITY_STATUS free_key(NCRYPT_HANDLE key)
{
    struct ncrypt_key_instance *keyInstance;
    struct ncrypt_provider_instance *providerInstance;
    SECURITY_STATUS ret;


    if (!key)
    {
        return NTE_INVALID_HANDLE;
    }

    keyInstance = handle2key(key);
    providerInstance = handle2provider(keyInstance->provider);


    if (keyInstance->kspHandle)
    {
        ret = providerInstance->functions.FreeKey(providerInstance->kspHandle, keyInstance->kspHandle);
        if (ret != ERROR_SUCCESS)
        {
            ERR("FreeKey failed in the key storage provider\n");
            return ret;
        }
        keyInstance->kspHandle = 0;
    }

    if (keyInstance->provider)
    {
        ret = free_provider(keyInstance->provider);
        if (ret != ERROR_SUCCESS)
        {
            return ret;
        }
        keyInstance->provider = 0;
    }

    heap_free(keyInstance);
    return ERROR_SUCCESS;
}
