#!/usr/bin/env python3
#
# Copyright 2021 Giovanni Mascellani for CodeWeavers
# Script to regenerate timezones data for Wine
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
#
# HOW TO USE THIS SCRIPT
#
# * Download the file windowsZones.xml from:
#     https://raw.githubusercontent.com/unicode-org/cldr/master/common/supplemental/windowsZones.xml
#   The same file might be packaged in your distribution (Debian has
#   it in unicode-cldr-core), but that copy might be outdated
#
# * Clone the TZ database from:
#     https://github.com/eggert/tz/
#   Make it and run:
#     mkdir out ; ./zic -d out -L /dev/null tzdata.zi
#   The directory "out" will have the binary TZ data files. The same
#   files will most probably be available in your distribution (Debian
#   has them in tzdata, where they correspond to the "posix" subdir,
#   not the "right" one), but again they might be outdated.
#
# * Call this script with (adjusting paths):
#     ./parse_windows_zones.py registry windowsZones.xml tz/out tz/backward | grep -v ^#
#   Paste the output to loader/wine.inf.in, deleting the corresponding old lines.
#
# * Call this script again with:
#     ./parse_windows_zones.py resource windowsZones.xml tz/out tz/backward | grep -v ^#
#   Paste the output to dlls/tzres/tzres.rc, deleting the corresponding old lines.
#
# Remember you can pipe stuff to "xclip -in -selection clipboard" to
# copy it to the X11 clipboard. Remove the "grep -v ^#" command if you
# want to see the debug output.

import re
import sys
import hashlib
import struct
import os
import datetime

COMMENT_RE = re.compile('<!-- +(\(UTC.*) -->.*')
MAP_RE = re.compile('<mapZone other="(.*)" territory=".*" type="(.*)"/>')

# Based on https://pubs.opengroup.org/onlinepubs/9699919799/ (variable TZ)
def read_std(footer):
    if footer[0] == '<':
        for i in range(len(footer)):
            if footer[i] == '>':
                return footer[1:i], footer[i+1:]
        raise Exception("Cannot find >")
    else:
        for i in range(len(footer)):
            if not('A' <= footer[i] <= 'Z' or 'a' <= footer[i] <= 'z'):
                break
        return footer[:i], footer[i:]

def read_number(footer):
    for i in range(len(footer)):
        if not('0' <= footer[i] <= '9'):
            break
    else:
        i = len(footer)
    return int(footer[:i]), footer[i:]

def read_time(footer):
    minutes = 0
    seconds = 0
    hours, footer = read_number(footer)
    if len(footer) > 0 and footer[0] == ':':
        footer = footer[1:]
        minutes, footer = read_number(footer)
        if len(footer) > 0 and footer[0] == ':':
            footer = footer[1:]
            seconds, footer = read_number(footer)
    return (seconds + minutes * 60 + hours * 3600), footer

def read_offset(footer):
    sign = '+'
    if footer[0] in ['+', '-']:
        sign = footer[0]
        footer = footer[1:]
    time, footer = read_time(footer)
    return ((-1 if sign == '-' else 1) * time), footer

def read_date(footer):
    if footer[0] == 'J':
        footer = footer[1:]
        num, footer = read_number(footer)
        return ('J', num), footer
    elif footer[0] == 'M':
        footer = footer[1:]
        month, footer = read_number(footer)
        assert footer[0] == '.'
        footer = footer[1:]
        week, footer = read_number(footer)
        assert footer[0] == '.'
        footer = footer[1:]
        day, footer = read_number(footer)
        return ('M', month, week, day), footer
    else:
        num, footer = read_number(footer)
        return ('D', num), footer

def fix_date_for_tzi(time, date):
    date = list(date)
    days = time // (24 * 60 * 60)
    date[3] = (date[3] + days) % 7
    if days >= 4:
        # Pass to the following week
        date[2] += 1
        if date[2] >= 5:
            # Pass to the following month
            if date[1] < 12:
                date[2] = 1
                date[1] += 1
            else:
                # Cannot pass to the following year, just take last
                # week in month
                date[2] = 5
    if days <= -4:
        # Pass to the previous week
        date[2] -= 1
        if date[2] <= 0:
            # Pass to the previous month
            if date[1] > 1:
                date[2] = 5
                date[1] -= 1
            else:
                # Cannot pass to the previous year, just take first
                # week in month
                date[2] = 1
    return time % (24 * 60 * 60), tuple(date)

def decode_footer(footer):
    #print(footer)
    std, footer = read_std(footer)
    #print(std)
    std_offset, footer = read_offset(footer)
    #print(std_offset)
    if footer != '':
        dst, footer = read_std(footer)
        #print(dst)
        if footer[0] != ',':
            dst_offset, footer = read_offset(footer)
        else:
            dst_offset = std_offset - 3600
        #print(dst_offset)
        assert footer[0] == ','
        footer = footer[1:]
        dst_date, footer = read_date(footer)
        #print(dst_date)
        if footer[0] == '/':
            footer = footer[1:]
            dst_time, footer = read_offset(footer)
        else:
            dst_time = 2 * 3600
        #print(dst_time)
        assert footer[0] == ','
        footer = footer[1:]
        std_date, footer = read_date(footer)
        #print(std_date)
        if len(footer) > 0 and footer[0] == '/':
            footer = footer[1:]
            std_time, footer = read_offset(footer)
        else:
            std_time = 2 * 3600
        #print(std_time)
        assert footer == ''
    else:
        dst = None
        dst_offset = None
        dst_date = None
        dst_time = None
        std_date = None
        std_time = None

    return std, std_offset, dst, dst_offset, dst_date, dst_time, std_date, std_time

# Create the TZI structure according to https://docs.microsoft.com/en-us/windows/win32/api/timezoneapi/ns-timezoneapi-time_zone_information
def footer_to_tzi(footer):
    std, std_offset, dst, dst_offset, dst_date, dst_time, std_date, std_time = footer
    if dst is not None:
        assert std_offset % 60 == 0
        assert dst_offset % 60 == 0
        tzi = struct.pack("<lll", std_offset // 60, 0, (dst_offset - std_offset) // 60)
        if std_date[0] == 'M':
            std_time, std_date = fix_date_for_tzi(std_time, std_date)
            tzi += struct.pack("<hhhhhhhh", 0, std_date[1], std_date[3], std_date[2], std_time // 3600, (std_time // 60) % 60, std_time % 60, 0)
        else:
            print('# Windows format cannot represent fixed dates')
            tzi += struct.pack("<hhhhhhhh", 0, 0, 0, 0, 0, 0, 0, 0)
        if dst_date[0] == 'M':
            dst_time, dst_date = fix_date_for_tzi(dst_time, dst_date)
            tzi += struct.pack("<hhhhhhhh", 0, dst_date[1], dst_date[3], dst_date[2], dst_time // 3600, (dst_time // 60) % 60, dst_time % 60, 0)
        else:
            print('# Windows format cannot represent fixed dates')
            tzi += struct.pack("<hhhhhhhh", 0, 0, 0, 0, 0, 0, 0, 0)
    else:
        assert std_offset % 60 == 0
        tzi = struct.pack("<lll", std_offset // 60, 0, -60)
        tzi += struct.pack("<hhhhhhhh", 0, 0, 0, 0, 0, 0, 0, 0)
        tzi += struct.pack("<hhhhhhhh", 0, 0, 0, 0, 0, 0, 0, 0)
    #print(",".join(["{:02x}".format(x) for x in tzi]))
    return tzi

def realize_date(year, month, weekday, day):
    diff = (weekday - datetime.date(year=year, month=month, day=1).weekday()) % 7
    actual_day = 1 + diff + 7 * (day-1)
    try:
        date = datetime.datetime(year=year, month=month, day=actual_day, hour=0, minute=0, second=0, microsecond=0)
    except ValueError:
        date = datetime.datetime(year=year, month=month, day=actual_day-7, hour=0, minute=0, second=0, microsecond=0)
    return date

def realize_footer_date(year, footer_date):
    if footer_date[0] == 'M':
        return realize_date(year, footer_date[1], (footer_date[3]-1) % 7, footer_date[2])

    elif footer_date[0] == 'J':
        # Use 2021 as the model non-leap year
        tmp = datetime.date(year=2021, month=1, day=1) + datetime.timedelta(days=footer_date[1]-1)
        return datetime.date(year=year, month=tmp.month, day=tmp.day)

    elif footer_date[0] == 'D':
        return datetime.date(year=year, month=1, day=1) + datetime.timedelta(days=footer_date[1])

    else:
        assert False

# def footer_to_tzi_for_year(footer, year):
#     std, std_offset, dst, dst_offset, dst_date, dst_time, std_date, std_time = footer
#     if dst is not None:
#         assert std_offset % 60 == 0
#         assert dst_offset % 60 == 0
#         tzi = struct.pack("<lll", std_offset // 60, 0, (dst_offset - std_offset) // 60)
#         actual_std_date = realize_footer_date(year, std_date)
#         actual_dst_date = realize_footer_date(year, dst_date)
#         tzi += struct.pack("<hhhhhhhh", 0, 0, 0, 0, 0, 0, 0, 0)
#         tzi += struct.pack("<hhhhhhhh", 0, 0, 0, 0, 0, 0, 0, 0)
#         #tzi += struct.pack("<hhhhhhhh", 0, std_date[1], std_date[3], std_date[2], std_time // 3600, (std_time // 60) % 60, std_time % 60, 0)
#         #tzi += struct.pack("<hhhhhhhh", 0, dst_date[1], dst_date[3], dst_date[2], dst_time // 3600, (dst_time // 60) % 60, dst_time % 60, 0)
#     else:
#         assert std_offset % 60 == 0
#         tzi = struct.pack("<lll", std_offset // 60, 0, -60)
#         tzi += struct.pack("<hhhhhhhh", 0, 0, 0, 0, 0, 0, 0, 0)
#         tzi += struct.pack("<hhhhhhhh", 0, 0, 0, 0, 0, 0, 0, 0)
#     return tzi

def realize_tzi(tzi, year):
    data = list(struct.unpack("<lllhhhhhhhhhhhhhhhh", tzi))
    for i in 0, 8:
        if data[3+i] == 0 and data[4+i] != 0:
            month = data[4+i]
            weekday = (data[5+i]-1) % 7
            day = data[6+i]
            hour = data[7+i]
            minute = data[8+i]
            second = data[9+i]
            millisecond = data[10+i]
            date = realize_date(year, month, weekday, day)
            date += datetime.timedelta(hours=hour, minutes=minute, seconds=second, microseconds=millisecond * 1000)
            data[3+i] = date.year
            data[4+i] = date.month
            data[5+i] = (date.weekday()+1) % 7
            data[6+i] = date.day
            data[7+i] = date.hour
            data[8+i] = date.minute
            data[9+i] = date.second
            data[10+i] = date.microsecond // 1000
    return struct.pack("<lllhhhhhhhhhhhhhhhh", *data)

# It mimicks std::lower_bound from C++, though it checks v[x][0], not
# v[x]
def lower_bound(v, x):
    low, high = 0, len(v)
    while low != high:
        i = (low + high) // 2
        if v[i][0] < x:
            low = i + 1
        else:
            high = i
    if low != len(v):
        assert not v[low][0] < x
    if low != 0:
        assert v[low-1][0] < x
    return low

def date_to_datetime(date):
    return datetime.datetime(year=date.year, month=date.month, day=date.day, hour=0, minute=0, second=0, microsecond=0)

FIRST_YEAR = 2000
LAST_YEAR = 2050

# Based on https://tools.ietf.org/html/rfc8536
def read_tz(filename):
    with open(filename, 'rb') as fin:
        header1 = fin.read(44)
        assert header1[:4] == b'TZif'
        assert header1[4:5] == b'2' or header1[4:5] == b'3'
        isutcnt, isstdcnt, leapcnt, timecnt, typecnt, charcnt = struct.unpack("!IIIIII", header1[20:])
        fin.read(timecnt * 4 + timecnt + typecnt * 6 + charcnt + leapcnt * 8 + isstdcnt + isutcnt)
        header2 = fin.read(44)
        assert header2[:4] == b'TZif'
        assert header2[4:5] == header1[4:5]
        isutcnt, isstdcnt, leapcnt, timecnt, typecnt, charcnt = struct.unpack("!IIIIII", header2[20:])

        transition_times = [struct.unpack("!q", fin.read(8))[0] for _ in range(timecnt)]
        transition_types = [struct.unpack("B", fin.read(1))[0] for _ in range(timecnt)]
        local_times = [struct.unpack("!i?B", fin.read(6)) for _ in range(typecnt)]
        designations = fin.read(charcnt).decode('ascii')
        leap_seconds = [struct.unpack("!Qi", fin.read(12)) for _ in range(leapcnt)]
        std_indicators = [struct.unpack("?", fin.read(1))[0] for _ in range(isstdcnt)]
        ut_indicators = [struct.unpack("?", fin.read(1))[0] for _ in range(isutcnt)]

        if std_indicators == []:
            std_indicators = [False] * typecnt
        if ut_indicators == []:
            ut_indicators = [False] * typecnt
        assert len(std_indicators) == len(local_times)
        assert len(ut_indicators) == len(local_times)
        assert len(leap_seconds) == 0

        print("# Local times:")
        descriptions = []
        utoffs = []
        dsts = []
        desigs = []
        for i in range(len(local_times)):
            utoff, dst, idx = local_times[i]
            for idx_end in range(idx, len(designations)):
                if designations[idx_end] == '\0':
                    break
            designation = designations[idx:idx_end]
            isstd = std_indicators[i]
            isut = ut_indicators[i]
            descr = "{}: {} (UT{:+}, {}) isstd={} isut={}".format(i+1, designation, utoff / 3600, 'daylight' if dst else 'standard', isstd, isut)
            descriptions.append(descr)
            utoffs.append(utoff)
            dsts.append(dst)
            desigs.append(designation)
            print("# {}".format(descr))
        print("# Transitions:")
        print("# Beginning {}".format(descriptions[0]))
        prev_tz = datetime.timezone(datetime.timedelta(seconds=utoffs[0]), name=desigs[0])
        prev_naive = None
        local_changes = []
        for i in range(len(transition_times)):
            change_utc = datetime.datetime.fromtimestamp(transition_times[i], tz=datetime.timezone.utc)
            change_prev = datetime.datetime.fromtimestamp(transition_times[i], tz=prev_tz)
            change_prev_naive = change_prev.replace(tzinfo=None)
            if prev_naive is not None:
                assert change_prev_naive > prev_naive
            local_changes.append((change_prev_naive, utoffs[transition_types[i]]))
            prev_naive = change_prev_naive
            print("# {} ({}, corresponding to {}) to {}".format(transition_times[i], change_utc, change_prev, descriptions[transition_types[i]]))
            prev_tz = datetime.timezone(datetime.timedelta(seconds=utoffs[transition_types[i]]), name=desigs[transition_types[i]])

        assert fin.read(1) == b'\n'
        footer = fin.readline().decode('ascii').strip()
        print("# Footer: {}".format(footer))
        footer_data = decode_footer(footer)
        std, std_offset, dst, dst_offset, dst_date, dst_time, std_date, std_time = footer_data
        tzi = footer_to_tzi(footer_data)

        # Complement local_changes with transitions generated with the
        # footer
        if dst is None:
            if local_changes == []:
                local_changes.append((datetime.datetime(year=FIRST_YEAR, month=1, day=1, hour=0, minute=0, second=0, microsecond=0), -std_offset))
        else:
            first_year = FIRST_YEAR
            if local_changes != [] and local_changes[-1][0].year < first_year:
                first_year = local_changes[-1][0].year
            for year in range(first_year, LAST_YEAR):
                start_of_year = datetime.datetime(year=year, month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
                actual_std_date = realize_footer_date(year, std_date)
                actual_dst_date = realize_footer_date(year, dst_date)
                change1 = (date_to_datetime(actual_std_date) + datetime.timedelta(seconds=std_time), -std_offset)
                change2 = (date_to_datetime(actual_dst_date) + datetime.timedelta(seconds=dst_time), -dst_offset)
                if change2[0] < change1[0]:
                    change1, change2 = change2, change1
                if local_changes == [] and change1[0] != start_of_year:
                    local_changes.append((start_of_year, change2[1]))
                if change1[0] > local_changes[-1][0]:
                    local_changes.append(change1)
                if change2[0] > local_changes[-1][0]:
                    local_changes.append(change2)

        # print("# Local changes:")
        # for change in local_changes:
        #     print("# {} to {}".format(change[0], change[1]))

        # For each year in a reasonable range, check if there is need
        # for a dynamic rule
        dynamic_tzis = []
        for year in range(FIRST_YEAR, LAST_YEAR):
            start_of_year = datetime.datetime(year=year, month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
            start_idx = lower_bound(local_changes, start_of_year)
            assert start_idx < len(local_changes) or dst is None
            if start_idx == len(local_changes):
                # Past the end of recorded data, so the footer is in
                # force
                pass
            else:
                if local_changes[start_idx][0] != start_of_year:
                    assert start_idx > 0
                    start_idx -= 1
                for end_idx in range(start_idx+1, len(local_changes)):
                    if local_changes[end_idx][0].year != year:
                        break
                else:
                    end_idx = len(local_changes)
                relevant_changes = []
                for i in range(start_idx, end_idx):
                    if i == start_idx or local_changes[i][1] != local_changes[i-1][1]:
                        relevant_changes.append(local_changes[i])

                if len(relevant_changes) > 3:
                    print("# WARNING! Too many changes in {}, will drop some of them".format(year))
                    relevant_changes = [relevant_changes[0], relevant_changes[1], relevant_changes[-1]]

                if len(relevant_changes) == 1:
                    [change] = relevant_changes
                    assert change[1] % 60 == 0
                    year_tzi = struct.pack("<lll", -change[1] // 60, 0, -60)
                    year_tzi += struct.pack("<hhhhhhhh", 0, 0, 0, 0, 0, 0, 0, 0)
                    year_tzi += struct.pack("<hhhhhhhh", 0, 0, 0, 0, 0, 0, 0, 0)

                elif len(relevant_changes) == 2:
                    [change1, change2] = relevant_changes
                    if change1[1] == change2[1]:
                        print("# WARNING! Not changing utoff in {}".format(year))
                    if change1[1] > change2[1]:
                        dst_change = change1
                        std_change = change2
                    else:
                        std_change = change1
                        dst_change = change2
                    if std_change[0].year != year:
                        std_change = (start_of_year, std_change[1])
                    if dst_change[0].year != year:
                        dst_change = (start_of_year, dst_change[1])
                    assert std_change[1] % 60 == 0
                    assert dst_change[1] % 60 == 0
                    std_week = (std_change[0].day-1) // 7 + 1
                    # if std_week == 4:
                    #     std_week = 5
                    dst_week = (dst_change[0].day-1) // 7 + 1
                    # if dst_week == 4:
                    #     dst_week = 5
                    year_tzi = struct.pack("<lll", -std_change[1] // 60, 0, -(dst_change[1] - std_change[1]) // 60)
                    year_tzi += struct.pack("<hhhhhhhh", 0, std_change[0].month, (std_change[0].weekday()+1) % 7, std_week,
                                            std_change[0].hour, std_change[0].minute, std_change[0].second, std_change[0].microsecond // 1000)
                    year_tzi += struct.pack("<hhhhhhhh", 0, dst_change[0].month, (dst_change[0].weekday()+1) % 7, dst_week,
                                            dst_change[0].hour, dst_change[0].minute, dst_change[0].second, dst_change[0].microsecond // 1000)

                elif len(relevant_changes) == 3:
                    [change1, change2, change3] = relevant_changes
                    if change1[1] == change2[1]:
                        print("# WARNING! Not changing utoff in {}".format(year))
                    if change2[1] == change3[1]:
                        print("# WARNING! Not changing utoff in {}".format(year))
                    if change1[1] != change3[1]:
                        print("# WARNING! Utoff doesn't switch back and forth in {}".format(year))
                    if change2[1] > change3[1]:
                        dst_change = change2
                        std_change = change3
                    else:
                        std_change = change2
                        dst_change = change3
                    assert std_change[1] % 60 == 0
                    assert dst_change[1] % 60 == 0
                    std_week = (std_change[0].day-1) // 7 + 1
                    # if std_week == 4:
                    #     std_week = 5
                    dst_week = (dst_change[0].day-1) // 7 + 1
                    # if dst_week == 4:
                    #     dst_week = 5
                    year_tzi = struct.pack("<lll", -std_change[1] // 60, 0, -(dst_change[1] - std_change[1]) // 60)
                    year_tzi += struct.pack("<hhhhhhhh", 0, std_change[0].month, (std_change[0].weekday()+1) % 7, std_week,
                                            std_change[0].hour, std_change[0].minute, std_change[0].second, std_change[0].microsecond // 1000)
                    year_tzi += struct.pack("<hhhhhhhh", 0, dst_change[0].month, (dst_change[0].weekday()+1) % 7, dst_week,
                                            dst_change[0].hour, dst_change[0].minute, dst_change[0].second, dst_change[0].microsecond // 1000)

                else:
                    assert(False)

                if realize_tzi(tzi, year) != realize_tzi(year_tzi, year):
                    dynamic_tzis.append((year, year_tzi))

        # If there are dynamic TZI entries at all, add another one at
        # the end with the plain tzi data, because Windows does the
        # same
        if dynamic_tzis != []:
            dynamic_tzis.append((dynamic_tzis[-1][0]+1, tzi))

        return tzi, dynamic_tzis

def main():
    command = sys.argv[1]
    windows_zones_filename = sys.argv[2]
    tzdata_dir = sys.argv[3]
    backward_filename = sys.argv[4]

    backward_data = {}
    with open(backward_filename) as fin:
        for line in fin:
            line = line.strip()
            if line == '' or line[0] == '#':
                continue
            fields = [x for x in line.split('\t') if x != '']
            assert len(fields) == 3
            assert fields[0] == 'Link'
            if fields[1] not in backward_data:
                backward_data[fields[1]] = set()
            backward_data[fields[1]].add(fields[2])

    display_names = {}
    olson_names = {}
    main_olson_name = {}
    current_display_name = None
    with open(windows_zones_filename) as fin:
        for line in fin:
            line = line.strip()
            match = COMMENT_RE.match(line)
            if match:
                #print("Display name: {}".format(match.group(1)))
                current_display_name = match.group(1)
                assert current_display_name not in olson_names
                assert current_display_name not in main_olson_name
                olson_names[current_display_name] = set()
                main_olson_name[current_display_name] = None
            match = MAP_RE.match(line)
            if match:
                #print("Registry name: {}; Olson name: {}".format(match.group(1), match.group(2)))
                reg_name = match.group(1)
                olson_name = match.group(2)
                if reg_name in display_names:
                    assert display_names[reg_name] == current_display_name
                else:
                    display_names[reg_name] = current_display_name
                for name in olson_name.split(' '):
                    olson_names[current_display_name].add(name)
                    if name in backward_data:
                        for name2 in backward_data[name]:
                            olson_names[current_display_name].add(name2)
                if main_olson_name[current_display_name] is None:
                    assert olson_name.split(' ') == [olson_name]
                    main_olson_name[current_display_name] = olson_name
    #print(reg_names)
    #print(olson_names)

    # Check that Olson names are not repeated
    all_olson_names = set()
    for olsons in olson_names.values():
        for name in olsons:
            if name in all_olson_names:
                assert False, "# WARNING! Olson name {} is repeated".format(name)
            all_olson_names.add(name)

    olson_seen = set()
    res_indices = set()
    for reg_name in sorted(display_names.keys(), key=str.lower):
        h = hashlib.sha1()
        h.update(reg_name.encode('ascii'))
        res_idx = int(h.hexdigest()[-3:], 16) << 4
        # Solve collisions by linear probing
        while res_idx in res_indices:
            res_idx += 16
        res_indices.add(res_idx)
        display_name = display_names[reg_name]
        dlt_name = reg_name.replace('Standard Time', 'Daylight Time')
        if command == 'registry':
            for olson_name in olson_names[display_name]:
                if olson_name in olson_seen:
                    print("# Olson name {} is duplicated".format(olson_name))
                olson_seen.add(olson_name)
            olson_name = main_olson_name[display_name]
            print("# Reading {} ({}, {})".format(olson_name, reg_name, display_name))
            tzi, dynamic_tzis = read_tz(os.path.join(tzdata_dir, olson_name))

            # See https://docs.microsoft.com/en-us/windows/win32/api/timezoneapi/ns-timezoneapi-time_zone_information
            print('HKLM,%CurrentVersionNT%\Time Zones\{},"Display",,"{}"'.format(reg_name, display_name))
            print('HKLM,%CurrentVersionNT%\Time Zones\{},"Dlt",,"{}"'.format(reg_name, dlt_name if dlt_name != 'UTC' else 'Coordinated Universal Time'))
            print('HKLM,%CurrentVersionNT%\Time Zones\{},"MUI_Display",,"@tzres.dll,-{}"'.format(reg_name, res_idx + 2))
            print('HKLM,%CurrentVersionNT%\Time Zones\{},"MUI_Dlt",,"@tzres.dll,-{}"'.format(reg_name, res_idx + 1))
            print('HKLM,%CurrentVersionNT%\Time Zones\{},"MUI_Std",,"@tzres.dll,-{}"'.format(reg_name, res_idx))
            print('HKLM,%CurrentVersionNT%\Time Zones\{},"Std",,"{}"'.format(reg_name, reg_name if reg_name != 'UTC' else 'Coordinated Universal Time'))
            print('HKLM,%CurrentVersionNT%\Time Zones\{},"TZI",1,{}'.format(reg_name, ",".join(["{:02x}".format(x) for x in tzi])))
            #print('HKLM,%CurrentVersionNT%\Time Zones\{},"Olson names",,"{}"'.format(reg_name, " ".join(sorted(olson_names[display_name]))))
            for year, year_tzi in dynamic_tzis:
                print('HKLM,%CurrentVersionNT%\Time Zones\{}\Dynamic DST,"{}",1,{}'.format(reg_name, year, ",".join(["{:02x}".format(x) for x in year_tzi])))
            if len(dynamic_tzis) != 0:
                print('HKLM,%CurrentVersionNT%\Time Zones\{}\Dynamic DST,"FirstEntry",0x10001,"{}"'.format(reg_name, dynamic_tzis[0][0]))
                print('HKLM,%CurrentVersionNT%\Time Zones\{}\Dynamic DST,"LastEntry",0x10001,"{}"'.format(reg_name, dynamic_tzis[-1][0]))
        if command == 'resource':
            print('{:7} "{}"'.format(res_idx, reg_name if reg_name != 'UTC' else 'Coordinated Universal Time'))
            print('{:7} "{}"'.format(res_idx + 1, dlt_name if dlt_name != 'UTC' else 'Coordinated Universal Time'))
            print('{:7} "{}"'.format(res_idx + 2, display_name))

if __name__ == '__main__':
    main()
    #read_tz("/etc/localtime")
